@@ -108,6 +108,12 @@ struct calibration_matrix {
 	float matrix[6];
 };
 
+enum input_config_mapped_to {
+	MAPPED_TO_DEFAULT,
+	MAPPED_TO_OUTPUT,
+	MAPPED_TO_REGION
+};
+
 /**
  * options for input devices
  */
@@ -147,7 +153,10 @@ struct input_config {
 	int xkb_capslock;
 
 	struct input_config_mapped_from_region *mapped_from_region;
+
+	enum input_config_mapped_to mapped_to;
 	char *mapped_to_output;
+	struct wlr_box *mapped_to_region;
 
 	bool capturable;
 	struct wlr_box region;
@@ -167,6 +176,21 @@ enum seat_config_allow_constrain {
 	CONSTRAIN_DISABLE
 };
 
+enum seat_keyboard_grouping {
+	KEYBOARD_GROUP_DEFAULT,  // the default is currently smart
+	KEYBOARD_GROUP_NONE,
+	KEYBOARD_GROUP_SMART  // keymap and repeat info
+};
+
+enum sway_input_idle_source {
+	IDLE_SOURCE_KEYBOARD = 1 << 0,
+	IDLE_SOURCE_POINTER = 1 << 1,
+	IDLE_SOURCE_TOUCH = 1 << 2,
+	IDLE_SOURCE_TABLET_PAD = 1 << 3,
+	IDLE_SOURCE_TABLET_TOOL = 1 << 4,
+	IDLE_SOURCE_SWITCH = 1 << 5,
+};
+
 /**
  * Options for multiseat and other misc device configurations
  */
@@ -176,6 +200,8 @@ struct seat_config {
 	list_t *attachments; // list of seat_attachment configs
 	int hide_cursor_timeout;
 	enum seat_config_allow_constrain allow_constrain;
+	enum seat_keyboard_grouping keyboard_grouping;
+	uint32_t idle_inhibit_sources, idle_wake_sources;
 	struct {
 		char *name;
 		int size;
@@ -188,6 +214,13 @@ enum config_dpms {
 	DPMS_OFF
 };
 
+enum scale_filter_mode {
+	SCALE_FILTER_DEFAULT, // the default is currently smart
+	SCALE_FILTER_LINEAR,
+	SCALE_FILTER_NEAREST,
+	SCALE_FILTER_SMART
+};
+
 /**
  * Size and position configuration for a particular output.
  *
@@ -198,10 +231,13 @@ struct output_config {
 	int enabled;
 	int width, height;
 	float refresh_rate;
+	int custom_mode;
 	int x, y;
 	float scale;
+	enum scale_filter_mode scale_filter;
 	int32_t transform;
 	enum wl_output_subpixel subpixel;
+	int max_render_time; // In milliseconds
 
 	char *background;
 	char *background_option;
@@ -333,8 +369,12 @@ enum edge_border_types {
 	E_VERTICAL,     /**< hide vertical edge borders */
 	E_HORIZONTAL,   /**< hide horizontal edge borders */
 	E_BOTH,		/**< hide vertical and horizontal edge borders */
-	E_SMART, /**< hide both if precisely one window is present in workspace */
-	E_SMART_NO_GAPS, /**< hide both if one window and gaps to edge is zero */
+};
+
+enum edge_border_smart_types {
+	ESMART_OFF,
+	ESMART_ON, /**< hide edges if precisely one window is present in workspace */
+	ESMART_NO_GAPS, /**< hide edges if one window and gaps to edge is zero */
 };
 
 enum sway_popup_during_fullscreen {
@@ -409,7 +449,8 @@ enum focus_follows_mouse_mode {
 enum focus_wrapping_mode {
 	WRAP_NO,
 	WRAP_YES,
-	WRAP_FORCE
+	WRAP_FORCE,
+	WRAP_WORKSPACE
 };
 
 enum mouse_warping_mode {
@@ -424,6 +465,12 @@ enum alignment {
 	ALIGN_RIGHT
 };
 
+enum xwayland_mode {
+	XWAYLAND_MODE_DISABLED,
+	XWAYLAND_MODE_LAZY,
+	XWAYLAND_MODE_IMMEDIATE
+};
+
 /**
  * The configuration struct. The result of loading a config file.
  */
@@ -464,7 +511,7 @@ struct sway_config {
 	size_t urgent_timeout;
 	enum sway_fowa focus_on_window_activation;
 	enum sway_popup_during_fullscreen popup_during_fullscreen;
-	bool xwayland;
+	enum xwayland_mode xwayland;
 
 	// swaybg
 	char *swaybg_command;
@@ -502,7 +549,7 @@ struct sway_config {
 	int border_thickness;
 	int floating_border_thickness;
 	enum edge_border_types hide_edge_borders;
-	enum edge_border_types saved_edge_borders;
+	enum edge_border_smart_types hide_edge_borders_smart;
 	bool hide_lone_tab;
 
 	// border colors
@@ -626,6 +673,8 @@ int output_name_cmp(const void *item, const void *data);
 void output_get_identifier(char *identifier, size_t len,
 	struct sway_output *output);
 
+const char *sway_output_scale_filter_to_string(enum scale_filter_mode scale_filter);
+
 struct output_config *new_output_config(const char *name);
 
 void merge_output_config(struct output_config *dst, struct output_config *src);
