@@ -54,6 +54,9 @@ struct sway_keyboard {
 	struct xkb_keymap *keymap;
 	xkb_layout_index_t effective_layout;
 
+	int32_t repeat_rate;
+	int32_t repeat_delay;
+
 	struct wl_listener keyboard_key;
 	struct wl_listener keyboard_modifiers;
 
@@ -67,6 +70,14 @@ struct sway_keyboard {
 	struct sway_binding *repeat_binding;
 };
 
+struct sway_keyboard_group {
+	struct wlr_keyboard_group *wlr_group;
+	struct sway_seat_device *seat_device;
+	struct wl_listener keyboard_key;
+	struct wl_listener keyboard_modifiers;
+	struct wl_list link; // sway_seat::keyboard_groups
+};
+
 struct xkb_keymap *sway_keyboard_compile_keymap(struct input_config *ic,
 		char **error);
 
