@@ -4,6 +4,7 @@
 #include <wlr/types/wlr_layer_shell_v1.h>
 #include <wlr/types/wlr_seat.h>
 #include <wlr/util/edges.h>
+#include "sway/config.h"
 #include "sway/input/input-manager.h"
 
 struct sway_seat;
@@ -28,6 +29,8 @@ struct sway_seat_device {
 	struct sway_input_device *input_device;
 	struct sway_keyboard *keyboard;
 	struct sway_switch *switch_device;
+	struct sway_tablet *tablet;
+	struct sway_tablet_pad *tablet_pad;
 	struct wl_list link; // sway_seat::devices
 };
 
@@ -78,6 +81,8 @@ struct sway_seat {
 
 	uint32_t last_button_serial;
 
+	uint32_t idle_inhibit_sources, idle_wake_sources;
+
 	list_t *deferred_bindings; // struct sway_binding
 
 	struct wl_listener focus_destroy;
@@ -88,6 +93,7 @@ struct sway_seat {
 	struct wl_listener request_set_primary_selection;
 
 	struct wl_list devices; // sway_seat_device::link
+	struct wl_list keyboard_groups; // sway_keyboard_group::link
 
 	struct wl_list link; // input_manager::seats
 };
@@ -193,6 +199,9 @@ struct seat_config *seat_get_config(struct sway_seat *seat);
 
 struct seat_config *seat_get_config_by_name(const char *name);
 
+void seat_idle_notify_activity(struct sway_seat *seat,
+		enum sway_input_idle_source source);
+
 bool seat_is_input_allowed(struct sway_seat *seat, struct wlr_surface *surface);
 
 void drag_icon_update_position(struct sway_drag_icon *icon);
