@@ -21,8 +21,8 @@ struct sway_layer_surface {
 	struct wl_listener output_destroy;
 	struct wl_listener new_popup;
 
-	bool configured;
 	struct wlr_box geo;
+	enum zwlr_layer_shell_v1_layer layer;
 };
 
 struct sway_layer_popup {
