@@ -32,6 +32,9 @@ struct sway_output {
 	int lx, ly; // layout coords
 	int width, height; // transformed buffer size
 	enum wl_output_subpixel detected_subpixel;
+	enum scale_filter_mode scale_filter;
+	// last applied mode when the output is DPMS'ed
+	struct wlr_output_mode *current_mode;
 
 	bool enabled, configured;
 	list_t *workspaces;
@@ -49,6 +52,12 @@ struct sway_output {
 	struct {
 		struct wl_signal destroy;
 	} events;
+
+	struct timespec last_presentation;
+	uint32_t refresh_nsec;
+	int max_render_time; // In milliseconds
+	struct wl_event_source *repaint_timer;
+	bool surface_needs_frame;
 };
 
 struct sway_output *output_create(struct wlr_output *wlr_output);
@@ -65,10 +74,12 @@ struct sway_output *output_get_in_direction(struct sway_output *reference,
 void output_add_workspace(struct sway_output *output,
 		struct sway_workspace *workspace);
 
-typedef void (*sway_surface_iterator_func_t)(struct sway_output *output,
+typedef void (*sway_surface_iterator_func_t)(struct sway_output *output, struct sway_view *view,
 	struct wlr_surface *surface, struct wlr_box *box, float rotation,
 	void *user_data);
 
+int output_repaint_timer_handler(void *data);
+
 void output_damage_whole(struct sway_output *output);
 
 void output_damage_surface(struct sway_output *output, double ox, double oy,
@@ -90,7 +101,7 @@ struct sway_output *all_output_by_name_or_id(const char *name_or_id);
 
 void output_sort_workspaces(struct sway_output *output);
 
-void output_enable(struct sway_output *output, struct output_config *oc);
+bool output_enable(struct sway_output *output, struct output_config *oc);
 
 void output_disable(struct sway_output *output);
 
