@@ -12,7 +12,6 @@
 #include <wlr/types/wlr_presentation_time.h>
 #include <wlr/types/wlr_relative_pointer_v1.h>
 #include <wlr/types/wlr_server_decoration.h>
-#include <wlr/types/wlr_xdg_shell_v6.h>
 #include <wlr/types/wlr_xdg_shell.h>
 #include "config.h"
 #include "list.h"
@@ -29,6 +28,8 @@ struct sway_server {
 	struct wlr_backend *noop_backend;
 
 	struct wlr_compositor *compositor;
+	struct wl_listener compositor_new_surface;
+
 	struct wlr_data_device_manager *data_device_manager;
 
 	struct sway_input_manager *input;
@@ -42,12 +43,11 @@ struct sway_server {
 	struct wlr_layer_shell_v1 *layer_shell;
 	struct wl_listener layer_shell_surface;
 
-	struct wlr_xdg_shell_v6 *xdg_shell_v6;
-	struct wl_listener xdg_shell_v6_surface;
-
 	struct wlr_xdg_shell *xdg_shell;
 	struct wl_listener xdg_shell_surface;
 
+	struct wlr_tablet_manager_v2 *tablet_v2;
+
 #if HAVE_XWAYLAND
 	struct sway_xwayland xwayland;
 	struct wl_listener xwayland_surface;
@@ -101,11 +101,11 @@ void server_fini(struct sway_server *server);
 bool server_start(struct sway_server *server);
 void server_run(struct sway_server *server);
 
+void handle_compositor_new_surface(struct wl_listener *listener, void *data);
 void handle_new_output(struct wl_listener *listener, void *data);
 
 void handle_idle_inhibitor_v1(struct wl_listener *listener, void *data);
 void handle_layer_shell_surface(struct wl_listener *listener, void *data);
-void handle_xdg_shell_v6_surface(struct wl_listener *listener, void *data);
 void handle_xdg_shell_surface(struct wl_listener *listener, void *data);
 #if HAVE_XWAYLAND
 void handle_xwayland_surface(struct wl_listener *listener, void *data);
