@@ -92,6 +92,11 @@ struct sway_container {
 	double width_fraction;
 	double height_fraction;
 
+	// The share of space of the parent container that all children occupy
+	// Used for doing the resize calculations
+	double child_total_width;
+	double child_total_height;
+
 	// These are in layout coordinates.
 	double content_x, content_y;
 	int content_width, content_height;
@@ -352,4 +357,6 @@ void container_raise_floating(struct sway_container *con);
 
 bool container_is_scratchpad_hidden(struct sway_container *con);
 
+bool container_is_scratchpad_hidden_or_child(struct sway_container *con);
+
 #endif
