@@ -2,7 +2,6 @@
 #define _SWAY_VIEW_H
 #include <wayland-server-core.h>
 #include <wlr/types/wlr_surface.h>
-#include <wlr/types/wlr_xdg_shell_v6.h>
 #include "config.h"
 #if HAVE_XWAYLAND
 #include <wlr/xwayland.h>
@@ -14,7 +13,6 @@ struct sway_container;
 struct sway_xdg_decoration;
 
 enum sway_view_type {
-	SWAY_VIEW_XDG_SHELL_V6,
 	SWAY_VIEW_XDG_SHELL,
 #if HAVE_XWAYLAND
 	SWAY_VIEW_XWAYLAND,
@@ -98,7 +96,6 @@ struct sway_view {
 	list_t *executed_criteria; // struct criteria *
 
 	union {
-		struct wlr_xdg_surface_v6 *wlr_xdg_surface_v6;
 		struct wlr_xdg_surface *wlr_xdg_surface;
 #if HAVE_XWAYLAND
 		struct wlr_xwayland_surface *wlr_xwayland_surface;
@@ -111,22 +108,8 @@ struct sway_view {
 	} events;
 
 	struct wl_listener surface_new_subsurface;
-};
 
-struct sway_xdg_shell_v6_view {
-	struct sway_view view;
-
-	struct wl_listener commit;
-	struct wl_listener request_move;
-	struct wl_listener request_resize;
-	struct wl_listener request_maximize;
-	struct wl_listener request_fullscreen;
-	struct wl_listener set_title;
-	struct wl_listener set_app_id;
-	struct wl_listener new_popup;
-	struct wl_listener map;
-	struct wl_listener unmap;
-	struct wl_listener destroy;
+	int max_render_time; // In milliseconds
 };
 
 struct sway_xdg_shell_view {
@@ -214,15 +197,6 @@ struct sway_subsurface {
 	struct wl_listener destroy;
 };
 
-struct sway_xdg_popup_v6 {
-	struct sway_view_child child;
-
-	struct wlr_xdg_surface_v6 *wlr_xdg_surface_v6;
-
-	struct wl_listener new_popup;
-	struct wl_listener destroy;
-};
-
 struct sway_xdg_popup {
 	struct sway_view_child child;
 
@@ -332,8 +306,6 @@ void view_child_destroy(struct sway_view_child *child);
 
 struct sway_view *view_from_wlr_xdg_surface(
 	struct wlr_xdg_surface *xdg_surface);
-struct sway_view *view_from_wlr_xdg_surface_v6(
-	struct wlr_xdg_surface_v6 *xdg_surface_v6);
 #if HAVE_XWAYLAND
 struct sway_view *view_from_wlr_xwayland_surface(
 	struct wlr_xwayland_surface *xsurface);
