@@ -11,10 +11,13 @@
 int wrap(int i, int max);
 
 /**
- * Given a string that represents an RGB(A) color, return a uint32_t
- * version of the color.
+ * Given a string that represents an RGB(A) color, result will be set to a
+ * uint32_t version of the color, as long as it is valid. If it is invalid,
+ * then false will be returned and result will be untouched.
  */
-uint32_t parse_color(const char *color);
+bool parse_color(const char *color, uint32_t *result);
+
+void color_to_rgba(float dest[static 4], uint32_t color);
 
 /**
  * Given a string that represents a boolean, return the boolean value. This
@@ -32,6 +35,6 @@ float parse_float(const char *value);
 
 const char *sway_wl_output_subpixel_to_string(enum wl_output_subpixel subpixel);
 
-bool set_cloexec(int fd, bool cloexec);
+bool sway_set_cloexec(int fd, bool cloexec);
 
 #endif
