@@ -119,6 +119,7 @@ static struct cmd_handler command_handlers[] = {
 	{ "kill", cmd_kill },
 	{ "layout", cmd_layout },
 	{ "mark", cmd_mark },
+	{ "max_render_time", cmd_max_render_time },
 	{ "move", cmd_move },
 	{ "nop", cmd_nop },
 	{ "opacity", cmd_opacity },
@@ -198,9 +199,6 @@ static void set_config_node(struct sway_node *node) {
 
 list_t *execute_command(char *_exec, struct sway_seat *seat,
 		struct sway_container *con) {
-	list_t *res_list = create_list();
-	char *exec = strdup(_exec);
-	char *head = exec;
 	char *cmd;
 	char matched_delim = ';';
 	list_t *views = NULL;
@@ -213,9 +211,16 @@ list_t *execute_command(char *_exec, struct sway_seat *seat,
 		}
 	}
 
+	char *exec = strdup(_exec);
+	char *head = exec;
+	list_t *res_list = create_list();
+
+	if (!res_list || !exec) {
+		return NULL;
+	}
+
 	config->handler_context.seat = seat;
 
-	head = exec;
 	do {
 		for (; isspace(*head); ++head) {}
 		// Extract criteria (valid for this command list only).
@@ -562,34 +567,3 @@ char *cmd_results_to_json(list_t *res_list) {
 	json_object_put(result_array);
 	return res;
 }
-
-/**
- * Check and add color to buffer.
- *
- * return error object, or NULL if color is valid.
- */
-struct cmd_results *add_color(char *buffer, const char *color) {
-	int len = strlen(color);
-	if (len != 7 && len != 9) {
-		return cmd_results_new(CMD_INVALID,
-				"Invalid color definition %s", color);
-	}
-	if (color[0] != '#') {
-		return cmd_results_new(CMD_INVALID,
-				"Invalid color definition %s", color);
-	}
-	for (int i = 1; i < len; ++i) {
-		if (!isxdigit(color[i])) {
-			return cmd_results_new(CMD_INVALID,
-					"Invalid color definition %s", color);
-		}
-	}
-	strcpy(buffer, color);
-	// add default alpha channel if color was defined without it
-	if (len == 7) {
-		buffer[7] = 'f';
-		buffer[8] = 'f';
-	}
-	buffer[9] = '\0';
-	return NULL;
-}
