@@ -11,7 +11,15 @@ struct cmd_results *cmd_xwayland(int argc, char **argv) {
 	}
 
 #ifdef HAVE_XWAYLAND
-	bool xwayland = parse_boolean(argv[0], true);
+	enum xwayland_mode xwayland;
+	if (strcmp(argv[0], "force") == 0) {
+		xwayland = XWAYLAND_MODE_IMMEDIATE;
+	} else if (parse_boolean(argv[0], true)) {
+		xwayland = XWAYLAND_MODE_LAZY;
+	} else {
+		xwayland = XWAYLAND_MODE_DISABLED;
+	}
+
 	if (config->reloading && config->xwayland != xwayland) {
 		return cmd_results_new(CMD_FAILURE,
 				"xwayland can only be enabled/disabled at launch");
