@@ -31,6 +31,7 @@
 #include "stringop.h"
 #include "list.h"
 #include "log.h"
+#include "util.h"
 
 struct sway_config *config = NULL;
 
@@ -192,13 +193,6 @@ static void destroy_removed_seats(struct sway_config *old_config,
 	}
 }
 
-static void set_color(float dest[static 4], uint32_t color) {
-	dest[0] = ((color >> 16) & 0xff) / 255.0;
-	dest[1] = ((color >> 8) & 0xff) / 255.0;
-	dest[2] = (color & 0xff) / 255.0;
-	dest[3] = 1.0;
-}
-
 static void config_defaults(struct sway_config *config) {
 	if (!(config->swaynag_command = strdup("swaynag"))) goto cleanup;
 	config->swaynag_config_errors = (struct swaynag_instance){0};
@@ -248,7 +242,7 @@ static void config_defaults(struct sway_config *config) {
 	config->font_height = 17; // height of monospace 10
 	config->urgent_timeout = 500;
 	config->popup_during_fullscreen = POPUP_SMART;
-	config->xwayland = true;
+	config->xwayland = XWAYLAND_MODE_LAZY;
 
 	config->titlebar_border_thickness = 1;
 	config->titlebar_h_padding = 5;
@@ -296,41 +290,41 @@ static void config_defaults(struct sway_config *config) {
 	config->border_thickness = 2;
 	config->floating_border_thickness = 2;
 	config->hide_edge_borders = E_NONE;
-	config->saved_edge_borders = E_NONE;
+	config->hide_edge_borders_smart = ESMART_OFF;
 	config->hide_lone_tab = false;
 
 	// border colors
-	set_color(config->border_colors.focused.border, 0x4C7899);
-	set_color(config->border_colors.focused.background, 0x285577);
-	set_color(config->border_colors.focused.text, 0xFFFFFFFF);
-	set_color(config->border_colors.focused.indicator, 0x2E9EF4);
-	set_color(config->border_colors.focused.child_border, 0x285577);
+	color_to_rgba(config->border_colors.focused.border, 0x4C7899FF);
+	color_to_rgba(config->border_colors.focused.background, 0x285577FF);
+	color_to_rgba(config->border_colors.focused.text, 0xFFFFFFFF);
+	color_to_rgba(config->border_colors.focused.indicator, 0x2E9EF4FF);
+	color_to_rgba(config->border_colors.focused.child_border, 0x285577FF);
 
-	set_color(config->border_colors.focused_inactive.border, 0x333333);
-	set_color(config->border_colors.focused_inactive.background, 0x5F676A);
-	set_color(config->border_colors.focused_inactive.text, 0xFFFFFFFF);
-	set_color(config->border_colors.focused_inactive.indicator, 0x484E50);
-	set_color(config->border_colors.focused_inactive.child_border, 0x5F676A);
+	color_to_rgba(config->border_colors.focused_inactive.border, 0x333333FF);
+	color_to_rgba(config->border_colors.focused_inactive.background, 0x5F676AFF);
+	color_to_rgba(config->border_colors.focused_inactive.text, 0xFFFFFFFF);
+	color_to_rgba(config->border_colors.focused_inactive.indicator, 0x484E50FF);
+	color_to_rgba(config->border_colors.focused_inactive.child_border, 0x5F676AFF);
 
-	set_color(config->border_colors.unfocused.border, 0x333333);
-	set_color(config->border_colors.unfocused.background, 0x222222);
-	set_color(config->border_colors.unfocused.text, 0x88888888);
-	set_color(config->border_colors.unfocused.indicator, 0x292D2E);
-	set_color(config->border_colors.unfocused.child_border, 0x222222);
+	color_to_rgba(config->border_colors.unfocused.border, 0x333333FF);
+	color_to_rgba(config->border_colors.unfocused.background, 0x222222FF);
+	color_to_rgba(config->border_colors.unfocused.text, 0x88888888);
+	color_to_rgba(config->border_colors.unfocused.indicator, 0x292D2EFF);
+	color_to_rgba(config->border_colors.unfocused.child_border, 0x222222FF);
 
-	set_color(config->border_colors.urgent.border, 0x2F343A);
-	set_color(config->border_colors.urgent.background, 0x900000);
-	set_color(config->border_colors.urgent.text, 0xFFFFFFFF);
-	set_color(config->border_colors.urgent.indicator, 0x900000);
-	set_color(config->border_colors.urgent.child_border, 0x900000);
+	color_to_rgba(config->border_colors.urgent.border, 0x2F343AFF);
+	color_to_rgba(config->border_colors.urgent.background, 0x900000FF);
+	color_to_rgba(config->border_colors.urgent.text, 0xFFFFFFFF);
+	color_to_rgba(config->border_colors.urgent.indicator, 0x900000FF);
+	color_to_rgba(config->border_colors.urgent.child_border, 0x900000FF);
 
-	set_color(config->border_colors.placeholder.border, 0x000000);
-	set_color(config->border_colors.placeholder.background, 0x0C0C0C);
-	set_color(config->border_colors.placeholder.text, 0xFFFFFFFF);
-	set_color(config->border_colors.placeholder.indicator, 0x000000);
-	set_color(config->border_colors.placeholder.child_border, 0x0C0C0C);
+	color_to_rgba(config->border_colors.placeholder.border, 0x000000FF);
+	color_to_rgba(config->border_colors.placeholder.background, 0x0C0C0CFF);
+	color_to_rgba(config->border_colors.placeholder.text, 0xFFFFFFFF);
+	color_to_rgba(config->border_colors.placeholder.indicator, 0x000000FF);
+	color_to_rgba(config->border_colors.placeholder.child_border, 0x0C0C0CFF);
 
-	set_color(config->border_colors.background, 0xFFFFFF);
+	color_to_rgba(config->border_colors.background, 0xFFFFFFFF);
 
 	// Security
 	if (!(config->command_policies = create_list())) goto cleanup;
@@ -447,15 +441,17 @@ bool load_main_config(const char *file, bool is_active, bool validating) {
 				old_config->xwayland ? "enabled" : "disabled");
 		config->xwayland = old_config->xwayland;
 
-		if (old_config->swaybg_client != NULL) {
-			wl_client_destroy(old_config->swaybg_client);
-		}
+		if (!config->validating) {
+			if (old_config->swaybg_client != NULL) {
+				wl_client_destroy(old_config->swaybg_client);
+			}
 
-		if (old_config->swaynag_config_errors.client != NULL) {
-			wl_client_destroy(old_config->swaynag_config_errors.client);
-		}
+			if (old_config->swaynag_config_errors.client != NULL) {
+				wl_client_destroy(old_config->swaynag_config_errors.client);
+			}
 
-		input_manager_reset_all_inputs();
+			input_manager_reset_all_inputs();
+		}
 	}
 
 	config->current_config_path = path;
@@ -520,8 +516,18 @@ bool load_main_config(const char *file, bool is_active, bool validating) {
 		return success;
 	}
 
-	if (is_active) {
+	if (is_active && !validating) {
 		input_manager_verify_fallback_seat();
+
+		for (int i = 0; i < config->input_configs->length; i++) {
+			input_manager_apply_input_config(config->input_configs->items[i]);
+		}
+
+		for (int i = 0; i < config->input_type_configs->length; i++) {
+			input_manager_apply_input_config(
+					config->input_type_configs->items[i]);
+		}
+
 		for (int i = 0; i < config->seat_configs->length; i++) {
 			input_manager_apply_seat_config(config->seat_configs->items[i]);
 		}
@@ -677,7 +683,7 @@ static ssize_t getline_with_cont(char **lineptr, size_t *line_size, FILE *file,
 	size_t next_line_size = 0;
 	ssize_t nread = getline(lineptr, line_size, file);
 	*nlines = nread == -1 ? 0 : 1;
-	while (nread >= 2 && strcmp(&(*lineptr)[nread - 2], "\\\n") == 0) {
+	while (nread >= 2 && strcmp(&(*lineptr)[nread - 2], "\\\n") == 0 && (*lineptr)[0] != '#') {
 		ssize_t next_nread = getline(&next_line, &next_line_size, file);
 		if (next_nread == -1) {
 			break;
@@ -687,8 +693,10 @@ static ssize_t getline_with_cont(char **lineptr, size_t *line_size, FILE *file,
 		nread += next_nread - 2;
 		if ((ssize_t) *line_size < nread + 1) {
 			*line_size = nread + 1;
+			char *old_ptr = *lineptr;
 			*lineptr = realloc(*lineptr, *line_size);
 			if (!*lineptr) {
+				free(old_ptr);
 				nread = -1;
 				break;
 			}
