@@ -29,6 +29,21 @@ void output_get_identifier(char *identifier, size_t len,
 		wlr_output->serial);
 }
 
+const char *sway_output_scale_filter_to_string(enum scale_filter_mode scale_filter) {
+	switch (scale_filter) {
+	case SCALE_FILTER_DEFAULT:
+		return "smart";
+	case SCALE_FILTER_LINEAR:
+		return "linear";
+	case SCALE_FILTER_NEAREST:
+		return "nearest";
+	case SCALE_FILTER_SMART:
+		return "smart";
+	}
+	sway_assert(false, "Unknown value for scale_filter.");
+	return NULL;
+}
+
 struct output_config *new_output_config(const char *name) {
 	struct output_config *oc = calloc(1, sizeof(struct output_config));
 	if (oc == NULL) {
@@ -42,10 +57,13 @@ struct output_config *new_output_config(const char *name) {
 	oc->enabled = -1;
 	oc->width = oc->height = -1;
 	oc->refresh_rate = -1;
+	oc->custom_mode = -1;
 	oc->x = oc->y = -1;
 	oc->scale = -1;
+	oc->scale_filter = SCALE_FILTER_DEFAULT;
 	oc->transform = -1;
 	oc->subpixel = WL_OUTPUT_SUBPIXEL_UNKNOWN;
+	oc->max_render_time = -1;
 	return oc;
 }
 
@@ -68,15 +86,24 @@ void merge_output_config(struct output_config *dst, struct output_config *src) {
 	if (src->scale != -1) {
 		dst->scale = src->scale;
 	}
+	if (src->scale_filter != SCALE_FILTER_DEFAULT) {
+		dst->scale_filter = src->scale_filter;
+	}
 	if (src->subpixel != WL_OUTPUT_SUBPIXEL_UNKNOWN) {
 		dst->subpixel = src->subpixel;
 	}
 	if (src->refresh_rate != -1) {
 		dst->refresh_rate = src->refresh_rate;
 	}
+	if (src->custom_mode != -1) {
+		dst->custom_mode = src->custom_mode;
+	}
 	if (src->transform != -1) {
 		dst->transform = src->transform;
 	}
+	if (src->max_render_time != -1) {
+		dst->max_render_time = src->max_render_time;
+	}
 	if (src->background) {
 		free(dst->background);
 		dst->background = strdup(src->background);
@@ -149,11 +176,12 @@ static void merge_id_on_name(struct output_config *oc) {
 			list_add(config->output_configs, ion_oc);
 			sway_log(SWAY_DEBUG, "Generated id on name output config \"%s\""
 				" (enabled: %d) (%dx%d@%fHz position %d,%d scale %f "
-				"transform %d) (bg %s %s) (dpms %d)", ion_oc->name,
-				ion_oc->enabled, ion_oc->width, ion_oc->height,
+				"transform %d) (bg %s %s) (dpms %d) (max render time: %d)",
+				ion_oc->name, ion_oc->enabled, ion_oc->width, ion_oc->height,
 				ion_oc->refresh_rate, ion_oc->x, ion_oc->y, ion_oc->scale,
 				ion_oc->transform, ion_oc->background,
-				ion_oc->background_option, ion_oc->dpms_state);
+				ion_oc->background_option, ion_oc->dpms_state,
+				ion_oc->max_render_time);
 		}
 	}
 	free(id_on_name);
@@ -193,20 +221,25 @@ struct output_config *store_output_config(struct output_config *oc) {
 	}
 
 	sway_log(SWAY_DEBUG, "Config stored for output %s (enabled: %d) (%dx%d@%fHz "
-		"position %d,%d scale %f subpixel %s transform %d) (bg %s %s) (dpms %d)",
+		"position %d,%d scale %f subpixel %s transform %d) (bg %s %s) (dpms %d) "
+		"(max render time: %d)",
 		oc->name, oc->enabled, oc->width, oc->height, oc->refresh_rate,
 		oc->x, oc->y, oc->scale, sway_wl_output_subpixel_to_string(oc->subpixel),
-		oc->transform, oc->background, oc->background_option, oc->dpms_state);
+		oc->transform, oc->background, oc->background_option, oc->dpms_state,
+		oc->max_render_time);
 
 	return oc;
 }
 
-static bool set_mode(struct wlr_output *output, int width, int height,
-		float refresh_rate) {
+static void set_mode(struct wlr_output *output, int width, int height,
+		float refresh_rate, bool custom) {
 	int mhz = (int)(refresh_rate * 1000);
-	if (wl_list_empty(&output->modes)) {
+
+	if (wl_list_empty(&output->modes) || custom) {
 		sway_log(SWAY_DEBUG, "Assigning custom mode to %s", output->name);
-		return wlr_output_set_custom_mode(output, width, height, mhz);
+		wlr_output_set_custom_mode(output, width, height,
+			refresh_rate > 0 ? mhz : 0);
+		return;
 	}
 
 	struct wlr_output_mode *mode, *best = NULL;
@@ -216,17 +249,62 @@ static bool set_mode(struct wlr_output *output, int width, int height,
 				best = mode;
 				break;
 			}
-			best = mode;
+			if (best == NULL || mode->refresh > best->refresh) {
+				best = mode;
+			}
 		}
 	}
 	if (!best) {
 		sway_log(SWAY_ERROR, "Configured mode for %s not available", output->name);
-		sway_log(SWAY_INFO, "Picking default mode instead");
-		best = wl_container_of(output->modes.prev, mode, link);
+		sway_log(SWAY_INFO, "Picking preferred mode instead");
+		best = wlr_output_preferred_mode(output);
 	} else {
 		sway_log(SWAY_DEBUG, "Assigning configured mode to %s", output->name);
 	}
-	return wlr_output_set_mode(output, best);
+	wlr_output_set_mode(output, best);
+}
+
+/* Some manufacturers hardcode the aspect-ratio of the output in the physical
+ * size field. */
+static bool phys_size_is_aspect_ratio(struct wlr_output *output) {
+	return (output->phys_width == 1600 && output->phys_height == 900) ||
+		(output->phys_width == 1600 && output->phys_height == 1000) ||
+		(output->phys_width == 160 && output->phys_height == 90) ||
+		(output->phys_width == 160 && output->phys_height == 100) ||
+		(output->phys_width == 16 && output->phys_height == 9) ||
+		(output->phys_width == 16 && output->phys_height == 10);
+}
+
+// The minimum DPI at which we turn on a scale of 2
+#define HIDPI_DPI_LIMIT (2 * 96)
+// The minimum screen height at which we turn on a scale of 2
+#define HIDPI_MIN_HEIGHT 1200
+// 1 inch = 25.4 mm
+#define MM_PER_INCH 25.4
+
+static int compute_default_scale(struct wlr_output *output) {
+	int width, height;
+	wlr_output_transformed_resolution(output, &width, &height);
+	if (height < HIDPI_MIN_HEIGHT) {
+		return 1;
+	}
+
+	if (output->phys_width == 0 || output->phys_height == 0) {
+		return 1;
+	}
+
+	if (phys_size_is_aspect_ratio(output)) {
+		return 1;
+	}
+
+	double dpi_x = (double) width / (output->phys_width / MM_PER_INCH);
+	double dpi_y = (double) height / (output->phys_height / MM_PER_INCH);
+	sway_log(SWAY_DEBUG, "Output DPI: %fx%f", dpi_x, dpi_y);
+	if (dpi_x <= HIDPI_DPI_LIMIT || dpi_y <= HIDPI_DPI_LIMIT) {
+		return 1;
+	}
+
+	return 2;
 }
 
 bool apply_output_config(struct output_config *oc, struct sway_output *output) {
@@ -238,51 +316,69 @@ bool apply_output_config(struct output_config *oc, struct sway_output *output) {
 
 	if (oc && !oc->enabled) {
 		// Output is configured to be disabled
+		sway_log(SWAY_DEBUG, "Disabling output %s", oc->name);
 		if (output->enabled) {
 			output_disable(output);
 			wlr_output_layout_remove(root->output_layout, wlr_output);
 		}
 		wlr_output_enable(wlr_output, false);
-		return true;
+		return wlr_output_commit(wlr_output);
 	} else if (!output->enabled) {
 		// Output is not enabled. Enable it, output_enable will call us again.
 		if (!oc || oc->dpms_state != DPMS_OFF) {
+			sway_log(SWAY_DEBUG, "Enabling output %s", oc->name);
 			wlr_output_enable(wlr_output, true);
+			wlr_output_commit(wlr_output);
 		}
-		output_enable(output, oc);
-		return true;
+		return output_enable(output, oc);
 	}
 
-	if (oc && oc->dpms_state == DPMS_ON) {
-		sway_log(SWAY_DEBUG, "Turning on screen");
+	if (!oc || oc->dpms_state != DPMS_OFF) {
+		sway_log(SWAY_DEBUG, "Turning on output %s", wlr_output->name);
 		wlr_output_enable(wlr_output, true);
+
+		if (oc && oc->width > 0 && oc->height > 0) {
+			sway_log(SWAY_DEBUG, "Set %s mode to %dx%d (%f Hz)", oc->name, oc->width,
+				oc->height, oc->refresh_rate);
+			set_mode(wlr_output, oc->width, oc->height,
+				oc->refresh_rate, oc->custom_mode == 1);
+		} else if (!wl_list_empty(&wlr_output->modes)) {
+			struct wlr_output_mode *mode = wlr_output_preferred_mode(wlr_output);
+			wlr_output_set_mode(wlr_output, mode);
+		}
+		output->current_mode = wlr_output->pending.mode;
 	}
 
-	bool modeset_success;
-	if (oc && oc->width > 0 && oc->height > 0) {
-		sway_log(SWAY_DEBUG, "Set %s mode to %dx%d (%f GHz)", oc->name, oc->width,
-			oc->height, oc->refresh_rate);
-		modeset_success =
-			set_mode(wlr_output, oc->width, oc->height, oc->refresh_rate);
-	} else if (!wl_list_empty(&wlr_output->modes)) {
-		struct wlr_output_mode *mode =
-			wl_container_of(wlr_output->modes.prev, mode, link);
-		modeset_success = wlr_output_set_mode(wlr_output, mode);
-	} else {
-		// Output doesn't support modes
-		modeset_success = true;
-	}
-	if (!modeset_success) {
-		// Failed to modeset, maybe the output is missing a CRTC. Leave the
-		// output disabled for now and try again when the output gets the mode
-		// we asked for.
-		sway_log(SWAY_ERROR, "Failed to modeset output %s", wlr_output->name);
-		return false;
-	}
-
+	float scale;
 	if (oc && oc->scale > 0) {
-		sway_log(SWAY_DEBUG, "Set %s scale to %f", oc->name, oc->scale);
-		wlr_output_set_scale(wlr_output, oc->scale);
+		scale = oc->scale;
+	} else {
+		scale = compute_default_scale(wlr_output);
+		sway_log(SWAY_DEBUG, "Auto-detected output scale: %f", scale);
+	}
+	if (scale != wlr_output->scale) {
+		sway_log(SWAY_DEBUG, "Set %s scale to %f", oc->name, scale);
+		wlr_output_set_scale(wlr_output, scale);
+	}
+
+	if (oc) {
+		enum scale_filter_mode scale_filter_old = output->scale_filter;
+		switch (oc->scale_filter) {
+			case SCALE_FILTER_DEFAULT:
+			case SCALE_FILTER_SMART:
+				output->scale_filter = ceilf(wlr_output->scale) == wlr_output->scale ?
+					SCALE_FILTER_NEAREST : SCALE_FILTER_LINEAR;
+				break;
+			case SCALE_FILTER_LINEAR:
+			case SCALE_FILTER_NEAREST:
+				output->scale_filter = oc->scale_filter;
+				break;
+		}
+		if (scale_filter_old != output->scale_filter) {
+			sway_log(SWAY_DEBUG, "Set %s scale_filter to %s", oc->name,
+				sway_output_scale_filter_to_string(output->scale_filter));
+			output_damage_whole(output);
+		}
 	}
 
 	if (oc && (oc->subpixel != WL_OUTPUT_SUBPIXEL_UNKNOWN || config->reloading)) {
@@ -297,6 +393,15 @@ bool apply_output_config(struct output_config *oc, struct sway_output *output) {
 		wlr_output_set_transform(wlr_output, oc->transform);
 	}
 
+	sway_log(SWAY_DEBUG, "Committing output %s", wlr_output->name);
+	if (!wlr_output_commit(wlr_output)) {
+		// Failed to modeset, maybe the output is missing a CRTC. Leave the
+		// output disabled for now and try again when the output gets the mode
+		// we asked for.
+		sway_log(SWAY_ERROR, "Failed to modeset output %s", wlr_output->name);
+		return false;
+	}
+
 	// Find position for it
 	if (oc && (oc->x != -1 || oc->y != -1)) {
 		sway_log(SWAY_DEBUG, "Set %s position to %d, %d", oc->name, oc->x, oc->y);
@@ -314,8 +419,15 @@ bool apply_output_config(struct output_config *oc, struct sway_output *output) {
 	output->height = output_box->height;
 
 	if (oc && oc->dpms_state == DPMS_OFF) {
-		sway_log(SWAY_DEBUG, "Turning off screen");
+		sway_log(SWAY_DEBUG, "Turning off output %s", oc->name);
 		wlr_output_enable(wlr_output, false);
+		wlr_output_commit(wlr_output);
+	}
+
+	if (oc && oc->max_render_time >= 0) {
+		sway_log(SWAY_DEBUG, "Set %s max render time to %d",
+			oc->name, oc->max_render_time);
+		output->max_render_time = oc->max_render_time;
 	}
 
 	return true;
@@ -324,19 +436,20 @@ bool apply_output_config(struct output_config *oc, struct sway_output *output) {
 static void default_output_config(struct output_config *oc,
 		struct wlr_output *wlr_output) {
 	oc->enabled = 1;
-	if (!wl_list_empty(&wlr_output->modes)) {
-		struct wlr_output_mode *mode =
-			wl_container_of(wlr_output->modes.prev, mode, link);
+	struct wlr_output_mode *mode = wlr_output_preferred_mode(wlr_output);
+	if (mode != NULL) {
 		oc->width = mode->width;
 		oc->height = mode->height;
-		oc->refresh_rate = mode->refresh;
+		oc->refresh_rate = mode->refresh / 1000.f;
 	}
 	oc->x = oc->y = -1;
-	oc->scale = 1;
+	oc->scale = 0; // auto
+	oc->scale_filter = SCALE_FILTER_DEFAULT;
 	struct sway_output *output = wlr_output->data;
 	oc->subpixel = output->detected_subpixel;
 	oc->transform = WL_OUTPUT_TRANSFORM_NORMAL;
 	oc->dpms_state = DPMS_ON;
+	oc->max_render_time = 0;
 }
 
 static struct output_config *get_output_config(char *identifier,
@@ -390,10 +503,11 @@ static struct output_config *get_output_config(char *identifier,
 
 		sway_log(SWAY_DEBUG, "Generated output config \"%s\" (enabled: %d)"
 			" (%dx%d@%fHz position %d,%d scale %f transform %d) (bg %s %s)"
-			" (dpms %d)", result->name, result->enabled, result->width,
-			result->height, result->refresh_rate, result->x, result->y,
-			result->scale, result->transform, result->background,
-			result->background_option, result->dpms_state);
+			" (dpms %d) (max render time: %d)", result->name, result->enabled,
+			result->width, result->height, result->refresh_rate,
+			result->x, result->y, result->scale, result->transform,
+			result->background, result->background_option, result->dpms_state,
+			result->max_render_time);
 	} else if (oc_name) {
 		// No identifier config, just return a copy of the name config
 		free(result->name);
@@ -436,8 +550,8 @@ void apply_output_config_to_outputs(struct output_config *oc) {
 	// will be applied during normal "new output" event from wlroots.
 	bool wildcard = strcmp(oc->name, "*") == 0;
 	char id[128];
-	struct sway_output *sway_output;
-	wl_list_for_each(sway_output, &root->all_outputs, link) {
+	struct sway_output *sway_output, *tmp;
+	wl_list_for_each_safe(sway_output, tmp, &root->all_outputs, link) {
 		char *name = sway_output->wlr_output->name;
 		output_get_identifier(id, sizeof(id), sway_output);
 		if (wildcard || !strcmp(name, oc->name) || !strcmp(id, oc->name)) {
@@ -505,7 +619,7 @@ static bool _spawn_swaybg(char **command) {
 		sway_log_errno(SWAY_ERROR, "socketpair failed");
 		return false;
 	}
-	if (!set_cloexec(sockets[0], true) || !set_cloexec(sockets[1], true)) {
+	if (!sway_set_cloexec(sockets[0], true) || !sway_set_cloexec(sockets[1], true)) {
 		return false;
 	}
 
@@ -529,7 +643,7 @@ static bool _spawn_swaybg(char **command) {
 			sway_log_errno(SWAY_ERROR, "fork failed");
 			_exit(EXIT_FAILURE);
 		} else if (pid == 0) {
-			if (!set_cloexec(sockets[1], false)) {
+			if (!sway_set_cloexec(sockets[1], false)) {
 				_exit(EXIT_FAILURE);
 			}
 
