@@ -200,7 +200,8 @@ void arrange_layers(struct sway_output *output) {
 	for (size_t i = 0; i < nlayers; ++i) {
 		wl_list_for_each_reverse(layer,
 				&output->layers[layers_above_shell[i]], link) {
-			if (layer->layer_surface->current.keyboard_interactive) {
+			if (layer->layer_surface->current.keyboard_interactive &&
+					layer->layer_surface->mapped) {
 				topmost = layer;
 				break;
 			}
@@ -214,6 +215,9 @@ void arrange_layers(struct sway_output *output) {
 	wl_list_for_each(seat, &server.input->seats, link) {
 		if (topmost != NULL) {
 			seat_set_focus_layer(seat, topmost->layer_surface);
+		} else if (seat->focused_layer &&
+				!seat->focused_layer->current.keyboard_interactive) {
+			seat_set_focus_layer(seat, NULL);
 		}
 	}
 }
@@ -276,7 +280,17 @@ static void handle_surface_commit(struct wl_listener *listener, void *data) {
 	struct sway_output *output = wlr_output->data;
 	struct wlr_box old_geo = layer->geo;
 	arrange_layers(output);
-	if (memcmp(&old_geo, &layer->geo, sizeof(struct wlr_box)) != 0) {
+
+	bool geo_changed =
+		memcmp(&old_geo, &layer->geo, sizeof(struct wlr_box)) != 0;
+	bool layer_changed = layer->layer != layer_surface->current.layer;
+	if (layer_changed) {
+		wl_list_remove(&layer->link);
+		wl_list_insert(&output->layers[layer_surface->current.layer],
+			&layer->link);
+		layer->layer = layer_surface->current.layer;
+	}
+	if (geo_changed || layer_changed) {
 		output_damage_surface(output, old_geo.x, old_geo.y,
 			layer_surface->surface, true);
 		output_damage_surface(output, layer->geo.x, layer->geo.y,
@@ -290,6 +304,15 @@ static void handle_surface_commit(struct wl_listener *listener, void *data) {
 }
 
 static void unmap(struct sway_layer_surface *sway_layer) {
+	struct sway_seat *seat;
+	wl_list_for_each(seat, &server.input->seats, link) {
+		if (seat->focused_layer == sway_layer->layer_surface) {
+			seat_set_focus_layer(seat, NULL);
+		}
+	}
+
+	cursor_rebase_all();
+
 	struct wlr_output *wlr_output = sway_layer->layer_surface->output;
 	if (wlr_output == NULL) {
 		return;
@@ -300,13 +323,6 @@ static void unmap(struct sway_layer_surface *sway_layer) {
 	}
 	output_damage_surface(output, sway_layer->geo.x, sway_layer->geo.y,
 		sway_layer->layer_surface->surface, true);
-
-	struct sway_seat *seat = input_manager_current_seat();
-	if (seat->focused_layer == sway_layer->layer_surface) {
-		seat_set_focus_layer(seat, NULL);
-	}
-
-	cursor_rebase_all();
 }
 
 static void handle_destroy(struct wl_listener *listener, void *data) {
@@ -370,10 +386,8 @@ static void popup_damage(struct sway_layer_popup *layer_popup, bool whole) {
 	while (true) {
 		if (layer_popup->parent_type == LAYER_PARENT_POPUP) {
 			layer_popup = layer_popup->parent_popup;
-			ox += layer_popup->wlr_popup->base->geometry.x +
-				layer_popup->wlr_popup->geometry.x;
-			oy += layer_popup->wlr_popup->base->geometry.y +
-				layer_popup->wlr_popup->geometry.y;
+			ox += layer_popup->wlr_popup->geometry.x;
+			oy += layer_popup->wlr_popup->geometry.y;
 		} else {
 			layer = layer_popup->parent_layer;
 			ox += layer->geo.x;
@@ -486,7 +500,9 @@ void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 	struct wlr_layer_surface_v1 *layer_surface = data;
 	sway_log(SWAY_DEBUG, "new layer surface: namespace %s layer %d anchor %d "
 			"size %dx%d margin %d,%d,%d,%d",
-		layer_surface->namespace, layer_surface->layer, layer_surface->layer,
+		layer_surface->namespace,
+		layer_surface->client_pending.layer,
+		layer_surface->client_pending.layer,
 		layer_surface->client_pending.desired_width,
 		layer_surface->client_pending.desired_height,
 		layer_surface->client_pending.margin.top,
@@ -543,7 +559,8 @@ void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 	sway_layer->output_destroy.notify = handle_output_destroy;
 	wl_signal_add(&output->events.destroy, &sway_layer->output_destroy);
 
-	wl_list_insert(&output->layers[layer_surface->layer], &sway_layer->link);
+	wl_list_insert(&output->layers[layer_surface->client_pending.layer],
+			&sway_layer->link);
 
 	// Temporarily set the layer's current state to client_pending
 	// So that we can easily arrange it
