@@ -23,6 +23,7 @@
 #include "sway/layers.h"
 #include "sway/output.h"
 #include "sway/server.h"
+#include "sway/surface.h"
 #include "sway/tree/arrange.h"
 #include "sway/tree/container.h"
 #include "sway/tree/root.h"
@@ -80,6 +81,7 @@ struct surface_iterator_data {
 	void *user_data;
 
 	struct sway_output *output;
+	struct sway_view *view;
 	double ox, oy;
 	int width, height;
 	float rotation;
@@ -134,7 +136,7 @@ static void output_for_each_surface_iterator(struct wlr_surface *surface,
 		return;
 	}
 
-	data->user_iterator(data->output, surface, &box, data->rotation,
+	data->user_iterator(data->output, data->view, surface, &box, data->rotation,
 		data->user_data);
 }
 
@@ -145,6 +147,7 @@ void output_surface_for_each_surface(struct sway_output *output,
 		.user_iterator = iterator,
 		.user_data = user_data,
 		.output = output,
+		.view = NULL,
 		.ox = ox,
 		.oy = oy,
 		.width = surface->current.width,
@@ -163,6 +166,7 @@ void output_view_for_each_surface(struct sway_output *output,
 		.user_iterator = iterator,
 		.user_data = user_data,
 		.output = output,
+		.view = view,
 		.ox = view->container->surface_x - output->lx
 			- view->geometry.x,
 		.oy = view->container->surface_y - output->ly
@@ -182,6 +186,7 @@ void output_view_for_each_popup(struct sway_output *output,
 		.user_iterator = iterator,
 		.user_data = user_data,
 		.output = output,
+		.view = view,
 		.ox = view->container->surface_x - output->lx
 			- view->geometry.x,
 		.oy = view->container->surface_y - output->ly
@@ -224,6 +229,7 @@ void output_layer_for_each_surface(struct sway_output *output,
 				.user_iterator = iterator,
 				.user_data = user_data,
 				.output = output,
+				.view = NULL,
 				.ox = popup_sx,
 				.oy = popup_sy,
 				.width = surface->current.width,
@@ -291,6 +297,7 @@ static void output_for_each_surface(struct sway_output *output,
 		.user_iterator = iterator,
 		.user_data = user_data,
 		.output = output,
+		.view = NULL,
 	};
 
 	struct sway_workspace *workspace = output_get_active_workspace(output);
@@ -375,7 +382,7 @@ bool output_has_opaque_overlay_layer_surface(struct sway_output *output) {
 	struct wlr_layer_surface_v1 *wlr_layer_surface_v1;
 	wl_list_for_each(wlr_layer_surface_v1, &server.layer_shell->surfaces, link) {
 		if (wlr_layer_surface_v1->output != output->wlr_output ||
-				wlr_layer_surface_v1->layer != ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY) {
+				wlr_layer_surface_v1->current.layer != ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY) {
 			continue;
 		}
 		struct wlr_surface *wlr_surface = wlr_layer_surface_v1->surface;
@@ -401,43 +408,234 @@ bool output_has_opaque_overlay_layer_surface(struct sway_output *output) {
 	return false;
 }
 
-static void send_frame_done_iterator(struct sway_output *output,
+struct send_frame_done_data {
+	struct timespec when;
+	int msec_until_refresh;
+};
+
+static void send_frame_done_iterator(struct sway_output *output, struct sway_view *view,
 		struct wlr_surface *surface, struct wlr_box *box, float rotation,
-		void *data) {
-	struct timespec *when = data;
-	wlr_surface_send_frame_done(surface, when);
-}
-
-static void send_frame_done(struct sway_output *output, struct timespec *when) {
-	output_for_each_surface(output, send_frame_done_iterator, when);
-}
-
-static void damage_handle_frame(struct wl_listener *listener, void *data) {
-	struct sway_output *output =
-		wl_container_of(listener, output, damage_frame);
-	if (!output->enabled || !output->wlr_output->enabled) {
-		return;
+		void *user_data) {
+	int view_max_render_time = 0;
+	if (view != NULL) {
+		view_max_render_time = view->max_render_time;
 	}
 
-	struct timespec now;
-	clock_gettime(CLOCK_MONOTONIC, &now);
+	struct send_frame_done_data *data = user_data;
+
+	int delay = data->msec_until_refresh - output->max_render_time
+			- view_max_render_time;
+
+	if (output->max_render_time == 0 || view_max_render_time == 0 || delay < 1) {
+		wlr_surface_send_frame_done(surface, &data->when);
+	} else {
+		struct sway_surface *sway_surface = surface->data;
+		wl_event_source_timer_update(sway_surface->frame_done_timer, delay);
+	}
+}
+
+static void send_frame_done(struct sway_output *output, struct send_frame_done_data *data) {
+	output_for_each_surface(output, send_frame_done_iterator, data);
+}
+
+static void count_surface_iterator(struct sway_output *output, struct sway_view *view,
+		struct wlr_surface *surface, struct wlr_box *_box, float rotation,
+		void *data) {
+	size_t *n = data;
+	(*n)++;
+}
+
+static bool scan_out_fullscreen_view(struct sway_output *output,
+		struct sway_view *view) {
+	struct wlr_output *wlr_output = output->wlr_output;
+	struct sway_workspace *workspace = output->current.active_workspace;
+	if (!sway_assert(workspace, "Expected an active workspace")) {
+		return false;
+	}
+
+	if (view->saved_buffer) {
+		return false;
+	}
+
+	for (int i = 0; i < workspace->current.floating->length; ++i) {
+		struct sway_container *floater =
+			workspace->current.floating->items[i];
+		if (container_is_transient_for(floater, view->container)) {
+			return false;
+		}
+	}
+
+#if HAVE_XWAYLAND
+	if (!wl_list_empty(&root->xwayland_unmanaged)) {
+		return false;
+	}
+#endif
+
+	if (!wl_list_empty(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY])) {
+		return false;
+	}
+	if (!wl_list_empty(&root->drag_icons)) {
+		return false;
+	}
+
+	struct wlr_surface *surface = view->surface;
+	if (surface == NULL) {
+		return false;
+	}
+	size_t n_surfaces = 0;
+	output_view_for_each_surface(output, view,
+		count_surface_iterator, &n_surfaces);
+	if (n_surfaces != 1) {
+		return false;
+	}
+
+	if (surface->buffer == NULL) {
+		return false;
+	}
+
+	if ((float)surface->current.scale != wlr_output->scale ||
+			surface->current.transform != wlr_output->transform) {
+		return false;
+	}
+
+	wlr_presentation_surface_sampled_on_output(server.presentation, surface,
+		wlr_output);
+
+	if (!wlr_output_attach_buffer(wlr_output, surface->buffer)) {
+		return false;
+	}
+	return wlr_output_commit(wlr_output);
+}
+
+int output_repaint_timer_handler(void *data) {
+	struct sway_output *output = data;
+	if (output->wlr_output == NULL) {
+		return 0;
+	}
+
+	output->wlr_output->frame_pending = false;
+
+	bool surface_needs_frame = output->surface_needs_frame;
+	output->surface_needs_frame = false;
+
+	struct sway_workspace *workspace = output->current.active_workspace;
+	if (workspace == NULL) {
+		return 0;
+	}
+
+	struct sway_container *fullscreen_con = root->fullscreen_global;
+	if (!fullscreen_con) {
+		fullscreen_con = workspace->current.fullscreen;
+	}
+
+	if (fullscreen_con && fullscreen_con->view) {
+		// Try to scan-out the fullscreen view
+		static bool last_scanned_out = false;
+		bool scanned_out =
+			scan_out_fullscreen_view(output, fullscreen_con->view);
+
+		if (scanned_out && !last_scanned_out) {
+			sway_log(SWAY_DEBUG, "Scanning out fullscreen view");
+		}
+		if (last_scanned_out && !scanned_out) {
+			sway_log(SWAY_DEBUG, "Stopping fullscreen view scan out");
+		}
+		last_scanned_out = scanned_out;
+
+		if (scanned_out) {
+			return 0;
+		}
+	}
 
 	bool needs_frame;
 	pixman_region32_t damage;
 	pixman_region32_init(&damage);
 	if (!wlr_output_damage_attach_render(output->damage,
 			&needs_frame, &damage)) {
-		return;
+		return 0;
 	}
 
 	if (needs_frame) {
+		struct timespec now;
+		clock_gettime(CLOCK_MONOTONIC, &now);
+
 		output_render(output, &now, &damage);
+	} else {
+		wlr_output_rollback(output->wlr_output);
+
+		if (surface_needs_frame) {
+			wlr_output_schedule_frame(output->wlr_output);
+		}
 	}
 
 	pixman_region32_fini(&damage);
 
+	return 0;
+}
+
+static void damage_handle_frame(struct wl_listener *listener, void *user_data) {
+	struct sway_output *output =
+		wl_container_of(listener, output, damage_frame);
+	if (!output->enabled || !output->wlr_output->enabled) {
+		return;
+	}
+
+	// Compute predicted milliseconds until the next refresh. It's used for
+	// delaying both output rendering and surface frame callbacks.
+	int msec_until_refresh = 0;
+
+	if (output->max_render_time != 0) {
+		struct timespec now;
+		clockid_t presentation_clock
+			= wlr_backend_get_presentation_clock(server.backend);
+		clock_gettime(presentation_clock, &now);
+
+		const long NSEC_IN_SECONDS = 1000000000;
+		struct timespec predicted_refresh = output->last_presentation;
+		predicted_refresh.tv_nsec += output->refresh_nsec % NSEC_IN_SECONDS;
+		predicted_refresh.tv_sec += output->refresh_nsec / NSEC_IN_SECONDS;
+		if (predicted_refresh.tv_nsec >= NSEC_IN_SECONDS) {
+			predicted_refresh.tv_sec += 1;
+			predicted_refresh.tv_nsec -= NSEC_IN_SECONDS;
+		}
+
+		// If the predicted refresh time is before the current time then
+		// there's no point in delaying.
+		//
+		// We only check tv_sec because if the predicted refresh time is less
+		// than a second before the current time, then msec_until_refresh will
+		// end up slightly below zero, which will effectively disable the delay
+		// without potential disastrous negative overflows that could occur if
+		// tv_sec was not checked.
+		if (predicted_refresh.tv_sec >= now.tv_sec) {
+			long nsec_until_refresh
+				= (predicted_refresh.tv_sec - now.tv_sec) * NSEC_IN_SECONDS
+					+ (predicted_refresh.tv_nsec - now.tv_nsec);
+
+			// We want msec_until_refresh to be conservative, that is, floored.
+			// If we have 7.9 msec until refresh, we better compute the delay
+			// as if we had only 7 msec, so that we don't accidentally delay
+			// more than necessary and miss a frame.
+			msec_until_refresh = nsec_until_refresh / 1000000;
+		}
+	}
+
+	int delay = msec_until_refresh - output->max_render_time;
+
+	// If the delay is less than 1 millisecond (which is the least we can wait)
+	// then just render right away.
+	if (delay < 1) {
+		output_repaint_timer_handler(output);
+	} else {
+		output->wlr_output->frame_pending = true;
+		wl_event_source_timer_update(output->repaint_timer, delay);
+	}
+
 	// Send frame done to all visible surfaces
-	send_frame_done(output, &now);
+	struct send_frame_done_data data = {0};
+	clock_gettime(CLOCK_MONOTONIC, &data.when);
+	data.msec_until_refresh = msec_until_refresh;
+	send_frame_done(output, &data);
 }
 
 void output_damage_whole(struct sway_output *output) {
@@ -448,7 +646,7 @@ void output_damage_whole(struct sway_output *output) {
 	}
 }
 
-static void damage_surface_iterator(struct sway_output *output,
+static void damage_surface_iterator(struct sway_output *output, struct sway_view *view,
 		struct wlr_surface *surface, struct wlr_box *_box, float rotation,
 		void *_data) {
 	bool *data = _data;
@@ -483,6 +681,7 @@ static void damage_surface_iterator(struct sway_output *output,
 		wlr_output_damage_add_box(output->damage, &box);
 	}
 
+	output->surface_needs_frame = true;
 	wlr_output_schedule_frame(output->wlr_output);
 }
 
@@ -566,6 +765,7 @@ static void update_output_manager_config(struct sway_server *server) {
 			root->output_layout, output->wlr_output);
 		// We mark the output enabled even if it is switched off by DPMS
 		config_head->state.enabled = output->enabled;
+		config_head->state.mode = output->current_mode;
 		if (output_box) {
 			config_head->state.x = output_box->x;
 			config_head->state.y = output_box->y;
@@ -653,14 +853,6 @@ static void handle_scale(struct wl_listener *listener, void *data) {
 	update_output_manager_config(output->server);
 }
 
-static void send_presented_iterator(struct sway_output *output,
-		struct wlr_surface *surface, struct wlr_box *box, float rotation,
-		void *data) {
-	struct wlr_presentation_event *event = data;
-	wlr_presentation_send_surface_presented(server.presentation,
-		surface, event);
-}
-
 static void handle_present(struct wl_listener *listener, void *data) {
 	struct sway_output *output = wl_container_of(listener, output, present);
 	struct wlr_output_event_present *output_event = data;
@@ -669,15 +861,8 @@ static void handle_present(struct wl_listener *listener, void *data) {
 		return;
 	}
 
-	struct wlr_presentation_event event = {
-		.output = output->wlr_output,
-		.tv_sec = (uint64_t)output_event->when->tv_sec,
-		.tv_nsec = (uint32_t)output_event->when->tv_nsec,
-		.refresh = (uint32_t)output_event->refresh,
-		.seq = (uint64_t)output_event->seq,
-		.flags = output_event->flags,
-	};
-	output_for_each_surface(output, send_presented_iterator, &event);
+	output->last_presentation = *output_event->when;
+	output->refresh_nsec = output_event->refresh;
 }
 
 void handle_new_output(struct wl_listener *listener, void *data) {
@@ -707,12 +892,16 @@ void handle_new_output(struct wl_listener *listener, void *data) {
 	wl_signal_add(&output->damage->events.destroy, &output->damage_destroy);
 	output->damage_destroy.notify = damage_handle_destroy;
 
+	output->repaint_timer = wl_event_loop_add_timer(server->wl_event_loop,
+		output_repaint_timer_handler, output);
+
 	struct output_config *oc = find_output_config(output);
 	if (!oc || oc->enabled) {
 		output_enable(output, oc);
 	} else {
 		wlr_output_enable(output->wlr_output, false);
 	}
+	free_output_config(oc);
 
 	transaction_commit_dirty();
 
@@ -760,7 +949,7 @@ void handle_output_manager_apply(struct wl_listener *listener, void *data) {
 			struct wlr_output_mode *mode = config_head->state.mode;
 			oc->width = mode->width;
 			oc->height = mode->height;
-			oc->refresh_rate = mode->refresh;
+			oc->refresh_rate = mode->refresh / 1000.f;
 		} else {
 			oc->width = config_head->state.custom_mode.width;
 			oc->height = config_head->state.custom_mode.height;
