@@ -1,9 +1,11 @@
 #define _POSIX_C_SOURCE 200809L
 #include <assert.h>
+#include <GLES2/gl2.h>
 #include <stdlib.h>
 #include <strings.h>
 #include <time.h>
 #include <wayland-server-core.h>
+#include <wlr/render/gles2.h>
 #include <wlr/render/wlr_renderer.h>
 #include <wlr/types/wlr_box.h>
 #include <wlr/types/wlr_buffer.h>
@@ -70,11 +72,39 @@ static void scissor_output(struct wlr_output *wlr_output,
 	wlr_renderer_scissor(renderer, &box);
 }
 
+static void set_scale_filter(struct wlr_output *wlr_output,
+		struct wlr_texture *texture, enum scale_filter_mode scale_filter) {
+	if (!wlr_texture_is_gles2(texture)) {
+		return;
+	}
+
+	struct wlr_gles2_texture_attribs attribs;
+	wlr_gles2_texture_get_attribs(texture, &attribs);
+
+	glBindTexture(attribs.target, attribs.tex);
+
+	switch (scale_filter) {
+	case SCALE_FILTER_LINEAR:
+		glTexParameteri(attribs.target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
+		break;
+	case SCALE_FILTER_NEAREST:
+		glTexParameteri(attribs.target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
+		break;
+	case SCALE_FILTER_DEFAULT:
+	case SCALE_FILTER_SMART:
+		assert(false); // unreachable
+	}
+}
+
 static void render_texture(struct wlr_output *wlr_output,
 		pixman_region32_t *output_damage, struct wlr_texture *texture,
 		const struct wlr_box *box, const float matrix[static 9], float alpha) {
 	struct wlr_renderer *renderer =
 		wlr_backend_get_renderer(wlr_output->backend);
+	struct sway_output *output = wlr_output->data;
+
+	struct wlr_gles2_texture_attribs attribs;
+	wlr_gles2_texture_get_attribs(texture, &attribs);
 
 	pixman_region32_t damage;
 	pixman_region32_init(&damage);
@@ -90,6 +120,7 @@ static void render_texture(struct wlr_output *wlr_output,
 	pixman_box32_t *rects = pixman_region32_rectangles(&damage, &nrects);
 	for (int i = 0; i < nrects; ++i) {
 		scissor_output(wlr_output, &rects[i]);
+		set_scale_filter(wlr_output, texture, output->scale_filter);
 		wlr_render_texture_with_matrix(renderer, texture, matrix, alpha);
 	}
 
@@ -97,7 +128,7 @@ damage_finish:
 	pixman_region32_fini(&damage);
 }
 
-static void render_surface_iterator(struct sway_output *output,
+static void render_surface_iterator(struct sway_output *output, struct sway_view *view,
 		struct wlr_surface *surface, struct wlr_box *_box, float rotation,
 		void *_data) {
 	struct render_data *data = _data;
@@ -120,6 +151,9 @@ static void render_surface_iterator(struct sway_output *output,
 		wlr_output->transform_matrix);
 
 	render_texture(wlr_output, output_damage, texture, &box, matrix, alpha);
+
+	wlr_presentation_surface_sampled_on_output(server.presentation, surface,
+		wlr_output);
 }
 
 static void render_layer(struct sway_output *output,
@@ -212,11 +246,11 @@ static void render_view_toplevels(struct sway_view *view,
 			render_surface_iterator, &data);
 }
 
-static void render_popup_iterator(struct sway_output *output,
+static void render_popup_iterator(struct sway_output *output, struct sway_view *view,
 		struct wlr_surface *surface, struct wlr_box *box, float rotation,
 		void *data) {
 	// Render this popup's surface
-	render_surface_iterator(output, surface, box, rotation, data);
+	render_surface_iterator(output, view, surface, box, rotation, data);
 
 	// Render this popup's child toplevels
 	output_surface_for_each_surface(output, surface, box->x, box->y,
@@ -267,6 +301,10 @@ static void render_saved_view(struct sway_view *view,
 
 	render_texture(wlr_output, damage, view->saved_buffer->texture,
 			&box, matrix, alpha);
+
+	// FIXME: we should set the surface that this saved buffer originates from
+	// as sampled here.
+	// https://github.com/swaywm/sway/pull/4465#discussion_r321082059
 }
 
 /**
@@ -937,6 +975,11 @@ void output_render(struct sway_output *output, struct timespec *when,
 		return;
 	}
 
+	struct sway_container *fullscreen_con = root->fullscreen_global;
+	if (!fullscreen_con) {
+		fullscreen_con = workspace->current.fullscreen;
+	}
+
 	wlr_renderer_begin(renderer, wlr_output->width, wlr_output->height);
 
 	if (!pixman_region32_not_empty(damage)) {
@@ -956,11 +999,6 @@ void output_render(struct sway_output *output, struct timespec *when,
 		goto render_overlay;
 	}
 
-	struct sway_container *fullscreen_con = root->fullscreen_global;
-	if (!fullscreen_con) {
-		fullscreen_con = workspace->current.fullscreen;
-	}
-
 	if (fullscreen_con) {
 		float clear_color[] = {0.0f, 0.0f, 0.0f, 1.0f};
 
