@@ -6,8 +6,10 @@
 #include <time.h>
 #include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_data_device.h>
+#include <wlr/types/wlr_idle.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_primary_selection.h>
+#include <wlr/types/wlr_tablet_v2.h>
 #include <wlr/types/wlr_xcursor_manager.h>
 #include "config.h"
 #include "list.h"
@@ -18,6 +20,7 @@
 #include "sway/input/keyboard.h"
 #include "sway/input/seat.h"
 #include "sway/input/switch.h"
+#include "sway/input/tablet.h"
 #include "sway/ipc-server.h"
 #include "sway/layers.h"
 #include "sway/output.h"
@@ -34,6 +37,8 @@ static void seat_device_destroy(struct sway_seat_device *seat_device) {
 	}
 
 	sway_keyboard_destroy(seat_device->keyboard);
+	sway_tablet_destroy(seat_device->tablet);
+	sway_tablet_pad_destroy(seat_device->tablet_pad);
 	wlr_cursor_detach_input_device(seat_device->sway_seat->cursor->cursor,
 		seat_device->input_device->wlr_device);
 	wl_list_remove(&seat_device->link);
@@ -67,6 +72,25 @@ static void seat_node_destroy(struct sway_seat_node *seat_node) {
 	free(seat_node);
 }
 
+void seat_idle_notify_activity(struct sway_seat *seat,
+		enum sway_input_idle_source source) {
+	uint32_t mask = seat->idle_inhibit_sources;
+	struct wlr_idle_timeout *timeout;
+	int ntimers = 0, nidle = 0;
+	wl_list_for_each(timeout, &server.idle->idle_timers, link) {
+		++ntimers;
+		if (timeout->idle_state) {
+			++nidle;
+		}
+	}
+	if (nidle == ntimers) {
+		mask = seat->idle_wake_sources;
+	}
+	if ((source & mask) > 0) {
+		wlr_idle_notify_activity(server.idle, seat->wlr_seat);
+	}
+}
+
 /**
  * Activate all views within this container recursively.
  */
@@ -98,6 +122,14 @@ static struct sway_keyboard *sway_keyboard_for_wlr_keyboard(
 			return seat_device->keyboard;
 		}
 	}
+	struct sway_keyboard_group *group;
+	wl_list_for_each(group, &seat->keyboard_groups, link) {
+		struct sway_input_device *input_device =
+			group->seat_device->input_device;
+		if (input_device->wlr_device->keyboard == wlr_keyboard) {
+			return group->seat_device->keyboard;
+		}
+	}
 	return NULL;
 }
 
@@ -118,6 +150,14 @@ static void seat_keyboard_notify_enter(struct sway_seat *seat,
 			state->pressed_keycodes, state->npressed, &keyboard->modifiers);
 }
 
+static void seat_tablet_pads_notify_enter(struct sway_seat *seat,
+		struct wlr_surface *surface) {
+	struct sway_seat_device *seat_device;
+	wl_list_for_each(seat_device, &seat->devices, link) {
+		sway_tablet_pad_notify_enter(seat_device->tablet_pad, surface);
+	}
+}
+
 /**
  * If con is a view, set it as active and enable keyboard input.
  * If con is a container, set all child views as active and don't enable
@@ -138,6 +178,7 @@ static void seat_send_focus(struct sway_node *node, struct sway_seat *seat) {
 #endif
 
 		seat_keyboard_notify_enter(seat, view->surface);
+		seat_tablet_pads_notify_enter(seat, view->surface);
 
 		struct wlr_pointer_constraint_v1 *constraint =
 			wlr_pointer_constraints_v1_constraint_for_surface(
@@ -470,9 +511,19 @@ struct sway_seat *seat_create(const char *seat_name) {
 		return NULL;
 	}
 
+	seat->idle_inhibit_sources = seat->idle_wake_sources =
+		IDLE_SOURCE_KEYBOARD |
+		IDLE_SOURCE_POINTER |
+		IDLE_SOURCE_TOUCH |
+		IDLE_SOURCE_TABLET_PAD |
+		IDLE_SOURCE_TABLET_TOOL |
+		IDLE_SOURCE_SWITCH;
+
 	// init the focus stack
 	wl_list_init(&seat->focus_stack);
 
+	wl_list_init(&seat->devices);
+
 	root_for_each_workspace(collect_focus_workspace_iter, seat);
 	root_for_each_container(collect_focus_container_iter, seat);
 
@@ -505,7 +556,7 @@ struct sway_seat *seat_create(const char *seat_name) {
 	seat->request_set_primary_selection.notify =
 		handle_request_set_primary_selection;
 
-	wl_list_init(&seat->devices);
+	wl_list_init(&seat->keyboard_groups);
 
 	wl_list_insert(&server.input->seats, &seat->link);
 
@@ -516,6 +567,7 @@ struct sway_seat *seat_create(const char *seat_name) {
 
 static void seat_update_capabilities(struct sway_seat *seat) {
 	uint32_t caps = 0;
+	uint32_t previous_caps = seat->wlr_seat->capabilities;
 	struct sway_seat_device *seat_device;
 	wl_list_for_each(seat_device, &seat->devices, link) {
 		switch (seat_device->input_device->wlr_device->type) {
@@ -545,7 +597,9 @@ static void seat_update_capabilities(struct sway_seat *seat) {
 		wlr_seat_set_capabilities(seat->wlr_seat, caps);
 	} else {
 		wlr_seat_set_capabilities(seat->wlr_seat, caps);
-		cursor_set_image(seat->cursor, "left_ptr", NULL);
+		if ((previous_caps & WL_SEAT_CAPABILITY_POINTER) == 0) {
+			cursor_set_image(seat->cursor, "left_ptr", NULL);
+		}
 	}
 }
 
@@ -559,41 +613,65 @@ static void seat_reset_input_config(struct sway_seat *seat,
 
 static void seat_apply_input_config(struct sway_seat *seat,
 		struct sway_seat_device *sway_device) {
-	const char *mapped_to_output = NULL;
-
-	struct input_config *ic = input_device_get_config(
-			sway_device->input_device);
-	if (ic != NULL) {
-		sway_log(SWAY_DEBUG, "Applying input config to %s",
-			sway_device->input_device->identifier);
-
-		mapped_to_output = ic->mapped_to_output;
+	struct input_config *ic =
+		input_device_get_config(sway_device->input_device);
+	if (ic == NULL) {
+		return;
 	}
 
-	if (mapped_to_output == NULL) {
+	sway_log(SWAY_DEBUG, "Applying input config to %s",
+		sway_device->input_device->identifier);
+
+	const char *mapped_to_output = ic->mapped_to_output;
+	struct wlr_box *mapped_to_region = ic->mapped_to_region;
+
+	switch (ic->mapped_to) {
+	case MAPPED_TO_DEFAULT:
 		mapped_to_output = sway_device->input_device->wlr_device->output_name;
-	}
-	if (mapped_to_output != NULL) {
+		if (mapped_to_output == NULL) {
+			return;
+		}
+		/* fallthrough */
+	case MAPPED_TO_OUTPUT:
 		sway_log(SWAY_DEBUG, "Mapping input device %s to output %s",
 			sway_device->input_device->identifier, mapped_to_output);
 		if (strcmp("*", mapped_to_output) == 0) {
 			wlr_cursor_map_input_to_output(seat->cursor->cursor,
 				sway_device->input_device->wlr_device, NULL);
+			wlr_cursor_map_input_to_region(seat->cursor->cursor,
+				sway_device->input_device->wlr_device, NULL);
 			sway_log(SWAY_DEBUG, "Reset output mapping");
 			return;
 		}
 		struct sway_output *output = output_by_name_or_id(mapped_to_output);
-		if (output) {
-			wlr_cursor_map_input_to_output(seat->cursor->cursor,
-				sway_device->input_device->wlr_device, output->wlr_output);
-			sway_log(SWAY_DEBUG, "Mapped to output %s", output->wlr_output->name);
+		if (!output) {
+			return;
 		}
+		wlr_cursor_map_input_to_output(seat->cursor->cursor,
+			sway_device->input_device->wlr_device, output->wlr_output);
+		wlr_cursor_map_input_to_region(seat->cursor->cursor,
+			sway_device->input_device->wlr_device, NULL);
+		sway_log(SWAY_DEBUG,
+			"Mapped to output %s", output->wlr_output->name);
+		return;
+	case MAPPED_TO_REGION:
+		sway_log(SWAY_DEBUG, "Mapping input device %s to %d,%d %dx%d",
+			sway_device->input_device->identifier,
+			mapped_to_region->x, mapped_to_region->y,
+			mapped_to_region->width, mapped_to_region->height);
+		wlr_cursor_map_input_to_output(seat->cursor->cursor,
+			sway_device->input_device->wlr_device, NULL);
+		wlr_cursor_map_input_to_region(seat->cursor->cursor,
+			sway_device->input_device->wlr_device, mapped_to_region);
+		return;
 	}
 }
 
 static void seat_configure_pointer(struct sway_seat *seat,
 		struct sway_seat_device *sway_device) {
-	seat_configure_xcursor(seat);
+	if ((seat->wlr_seat->capabilities & WL_SEAT_CAPABILITY_POINTER) == 0) {
+		seat_configure_xcursor(seat);
+	}
 	wlr_cursor_attach_input_device(seat->cursor->cursor,
 		sway_device->input_device->wlr_device);
 	seat_apply_input_config(seat, sway_device);
@@ -633,12 +711,23 @@ static void seat_configure_touch(struct sway_seat *seat,
 
 static void seat_configure_tablet_tool(struct sway_seat *seat,
 		struct sway_seat_device *sway_device) {
-	seat_configure_xcursor(seat);
+	if (!sway_device->tablet) {
+		sway_device->tablet = sway_tablet_create(seat, sway_device);
+	}
+	sway_configure_tablet(sway_device->tablet);
 	wlr_cursor_attach_input_device(seat->cursor->cursor,
 		sway_device->input_device->wlr_device);
 	seat_apply_input_config(seat, sway_device);
 }
 
+static void seat_configure_tablet_pad(struct sway_seat *seat,
+		struct sway_seat_device *sway_device) {
+	if (!sway_device->tablet) {
+		sway_device->tablet_pad = sway_tablet_pad_create(seat, sway_device);
+	}
+	sway_configure_tablet_pad(sway_device->tablet_pad);
+}
+
 static struct sway_seat_device *seat_get_device(struct sway_seat *seat,
 		struct sway_input_device *input_device) {
 	struct sway_seat_device *seat_device = NULL;
@@ -675,7 +764,7 @@ void seat_configure_device(struct sway_seat *seat,
 			seat_configure_tablet_tool(seat, seat_device);
 			break;
 		case WLR_INPUT_DEVICE_TABLET_PAD:
-			sway_log(SWAY_DEBUG, "TODO: configure tablet pad");
+			seat_configure_tablet_pad(seat, seat_device);
 			break;
 	}
 }
@@ -876,12 +965,9 @@ static void set_workspace(struct sway_seat *seat,
 
 	if (seat->workspace) {
 		free(seat->prev_workspace_name);
-		seat->prev_workspace_name = malloc(strlen(seat->workspace->name) + 1);
+		seat->prev_workspace_name = strdup(seat->workspace->name);
 		if (!seat->prev_workspace_name) {
 			sway_log(SWAY_ERROR, "Unable to allocate previous workspace name");
-			seat->prev_workspace_name = NULL;
-		} else {
-			strcpy(seat->prev_workspace_name, seat->workspace->name);
 		}
 	}
 
@@ -1072,6 +1158,7 @@ void seat_set_focus_surface(struct sway_seat *seat,
 		seat->has_focus = false;
 	}
 	seat_keyboard_notify_enter(seat, surface);
+	seat_tablet_pads_notify_enter(seat, surface);
 }
 
 void seat_set_focus_layer(struct sway_seat *seat,
@@ -1088,8 +1175,9 @@ void seat_set_focus_layer(struct sway_seat *seat,
 	} else if (!layer || seat->focused_layer == layer) {
 		return;
 	}
+	assert(layer->mapped);
 	seat_set_focus_surface(seat, layer->surface, true);
-	if (layer->layer >= ZWLR_LAYER_SHELL_V1_LAYER_TOP) {
+	if (layer->current.layer >= ZWLR_LAYER_SHELL_V1_LAYER_TOP) {
 		seat->focused_layer = layer;
 	}
 }
@@ -1265,6 +1353,9 @@ void seat_apply_config(struct sway_seat *seat,
 		return;
 	}
 
+	seat->idle_inhibit_sources = seat_config->idle_inhibit_sources;
+	seat->idle_wake_sources = seat_config->idle_wake_sources;
+
 	wl_list_for_each(seat_device, &seat->devices, link) {
 		seat_configure_device(seat, seat_device->input_device);
 	}
