@@ -7,6 +7,9 @@
 #include "sway/input/seat.h"
 #include "sway/tree/view.h"
 #include "log.h"
+#if HAVE_XWAYLAND
+#include "sway/xwayland.h"
+#endif
 
 struct seatop_default_event {
 	struct sway_node *previous_node;
@@ -58,8 +61,8 @@ static bool edge_is_external(struct sway_container *cont, enum wlr_edges edge) {
 }
 
 static enum wlr_edges find_edge(struct sway_container *cont,
-		struct sway_cursor *cursor) {
-	if (!cont->view) {
+		struct wlr_surface *surface, struct sway_cursor *cursor) {
+	if (!cont->view || (surface && cont->view->surface != surface)) {
 		return WLR_EDGE_NONE;
 	}
 	if (cont->border == B_NONE || !cont->border_thickness ||
@@ -89,8 +92,8 @@ static enum wlr_edges find_edge(struct sway_container *cont,
  * Edges that can't be resized are edges of the workspace.
  */
 static enum wlr_edges find_resize_edge(struct sway_container *cont,
-		struct sway_cursor *cursor) {
-	enum wlr_edges edge = find_edge(cont, cursor);
+		struct wlr_surface *surface, struct sway_cursor *cursor) {
+	enum wlr_edges edge = find_edge(cont, surface, cursor);
 	if (edge && !container_is_floating(cont) && edge_is_external(cont, edge)) {
 		return WLR_EDGE_NONE;
 	}
@@ -199,7 +202,7 @@ static void cursor_do_rebase(struct sway_cursor *cursor, uint32_t time_msec,
 		}
 	} else if (node && node->type == N_CONTAINER) {
 		// Try a node's resize edge
-		enum wlr_edges edge = find_resize_edge(node->sway_container, cursor);
+		enum wlr_edges edge = find_resize_edge(node->sway_container, surface, cursor);
 		if (edge == WLR_EDGE_NONE) {
 			cursor_set_image(cursor, "left_ptr", NULL);
 		} else if (container_is_floating(node->sway_container)) {
@@ -235,14 +238,15 @@ static void handle_button(struct sway_seat *seat, uint32_t time_msec,
 	double sx, sy;
 	struct sway_node *node = node_at_coords(seat,
 			cursor->cursor->x, cursor->cursor->y, &surface, &sx, &sy);
+
 	struct sway_container *cont = node && node->type == N_CONTAINER ?
 		node->sway_container : NULL;
 	bool is_floating = cont && container_is_floating(cont);
 	bool is_floating_or_child = cont && container_is_floating_or_child(cont);
 	bool is_fullscreen_or_child = cont && container_is_fullscreen_or_child(cont);
-	enum wlr_edges edge = cont ? find_edge(cont, cursor) : WLR_EDGE_NONE;
+	enum wlr_edges edge = cont ? find_edge(cont, surface, cursor) : WLR_EDGE_NONE;
 	enum wlr_edges resize_edge = cont && edge ?
-		find_resize_edge(cont, cursor) : WLR_EDGE_NONE;
+		find_resize_edge(cont, surface, cursor) : WLR_EDGE_NONE;
 	bool on_border = edge != WLR_EDGE_NONE;
 	bool on_contents = cont && !on_border && surface;
 	bool on_workspace = node && node->type == N_WORKSPACE;
@@ -335,10 +339,11 @@ static void handle_button(struct sway_seat *seat, uint32_t time_msec,
 			state == WLR_BUTTON_PRESSED) {
 		uint32_t btn_move = config->floating_mod_inverse ? BTN_RIGHT : BTN_LEFT;
 		if (button == btn_move && (mod_pressed || on_titlebar)) {
+			seat_set_focus_container(seat,
+					seat_get_focus_inactive_view(seat, &cont->node));
 			while (cont->parent) {
 				cont = cont->parent;
 			}
-			seat_set_focus_container(seat, cont);
 			seatop_begin_move_floating(seat, cont);
 			return;
 		}
@@ -400,13 +405,29 @@ static void handle_button(struct sway_seat *seat, uint32_t time_msec,
 	}
 
 	// Handle clicking a container surface or decorations
-	if (cont) {
+	if (cont && state == WLR_BUTTON_PRESSED) {
 		node = seat_get_focus_inactive(seat, &cont->node);
 		seat_set_focus(seat, node);
 		seat_pointer_notify_button(seat, time_msec, button, state);
 		return;
 	}
 
+#if HAVE_XWAYLAND
+	// Handle clicking on xwayland unmanaged view
+	if (surface && wlr_surface_is_xwayland_surface(surface)) {
+		struct wlr_xwayland_surface *xsurface =
+			wlr_xwayland_surface_from_wlr_surface(surface);
+		if (xsurface->override_redirect &&
+				wlr_xwayland_or_surface_wants_focus(xsurface)) {
+			struct wlr_xwayland *xwayland = server.xwayland.wlr_xwayland;
+			wlr_xwayland_set_seat(xwayland, seat->wlr_seat);
+			seat_set_focus_surface(seat, xsurface->surface, false);
+			seat_pointer_notify_button(seat, time_msec, button, state);
+			return;
+		}
+	}
+#endif
+
 	seat_pointer_notify_button(seat, time_msec, button, state);
 }
 
@@ -505,7 +526,7 @@ static void handle_axis(struct sway_seat *seat,
 			cursor->cursor->x, cursor->cursor->y, &surface, &sx, &sy);
 	struct sway_container *cont = node && node->type == N_CONTAINER ?
 		node->sway_container : NULL;
-	enum wlr_edges edge = cont ? find_edge(cont, cursor) : WLR_EDGE_NONE;
+	enum wlr_edges edge = cont ? find_edge(cont, surface, cursor) : WLR_EDGE_NONE;
 	bool on_border = edge != WLR_EDGE_NONE;
 	bool on_titlebar = cont && !on_border && !surface;
 	bool on_titlebar_border = cont && on_border &&
