@@ -163,7 +163,7 @@ static void ipc_json_describe_output(struct sway_output *output,
 	json_object_object_add(object, "type", json_object_new_string("output"));
 	json_object_object_add(object, "active", json_object_new_boolean(true));
 	json_object_object_add(object, "dpms",
-			json_object_new_boolean(output->wlr_output->enabled));
+			json_object_new_boolean(wlr_output->enabled));
 	json_object_object_add(object, "primary", json_object_new_boolean(false));
 	json_object_object_add(object, "layout", json_object_new_string("output"));
 	json_object_object_add(object, "orientation",
@@ -177,6 +177,9 @@ static void ipc_json_describe_output(struct sway_output *output,
 			json_object_new_string(wlr_output->serial));
 	json_object_object_add(object, "scale",
 			json_object_new_double(wlr_output->scale));
+	json_object_object_add(object, "scale_filter",
+		json_object_new_string(
+			sway_output_scale_filter_to_string(output->scale_filter)));
 	json_object_object_add(object, "transform",
 		json_object_new_string(
 			ipc_json_output_transform_description(wlr_output->transform)));
@@ -224,6 +227,8 @@ static void ipc_json_describe_output(struct sway_output *output,
 				* ((double)output->height / parent_box.height);
 		json_object_object_add(object, "percent", json_object_new_double(percent));
 	}
+
+	json_object_object_add(object, "max_render_time", json_object_new_int(output->max_render_time));
 }
 
 json_object *ipc_json_describe_disabled_output(struct sway_output *output) {
@@ -416,6 +421,8 @@ static void ipc_json_describe_view(struct sway_container *c, json_object *object
 	struct wlr_box geometry = {0, 0, c->view->natural_width, c->view->natural_height};
 	json_object_object_add(object, "geometry", ipc_json_create_rect(&geometry));
 
+	json_object_object_add(object, "max_render_time", json_object_new_int(c->view->max_render_time));
+
 #if HAVE_XWAYLAND
 	if (c->view->type == SWAY_VIEW_XWAYLAND) {
 		json_object_object_add(object, "window",
