@@ -19,6 +19,7 @@
 #include <wlr/types/wlr_relative_pointer_v1.h>
 #include <wlr/types/wlr_screencopy_v1.h>
 #include <wlr/types/wlr_server_decoration.h>
+#include <wlr/types/wlr_tablet_v2.h>
 #include <wlr/types/wlr_xcursor_manager.h>
 #include <wlr/types/wlr_xdg_decoration_v1.h>
 #include <wlr/types/wlr_xdg_output_v1.h>
@@ -58,6 +59,10 @@ bool server_init(struct sway_server *server) {
 	wlr_renderer_init_wl_display(renderer, server->wl_display);
 
 	server->compositor = wlr_compositor_create(server->wl_display, renderer);
+	server->compositor_new_surface.notify = handle_compositor_new_surface;
+	wl_signal_add(&server->compositor->events.new_surface,
+		&server->compositor_new_surface);
+
 	server->data_device_manager =
 		wlr_data_device_manager_create(server->wl_display);
 
@@ -81,16 +86,13 @@ bool server_init(struct sway_server *server) {
 		&server->layer_shell_surface);
 	server->layer_shell_surface.notify = handle_layer_shell_surface;
 
-	server->xdg_shell_v6 = wlr_xdg_shell_v6_create(server->wl_display);
-	wl_signal_add(&server->xdg_shell_v6->events.new_surface,
-		&server->xdg_shell_v6_surface);
-	server->xdg_shell_v6_surface.notify = handle_xdg_shell_v6_surface;
-
 	server->xdg_shell = wlr_xdg_shell_create(server->wl_display);
 	wl_signal_add(&server->xdg_shell->events.new_surface,
 		&server->xdg_shell_surface);
 	server->xdg_shell_surface.notify = handle_xdg_shell_surface;
 
+	server->tablet_v2 = wlr_tablet_v2_create(server->wl_display);
+
 	server->server_decoration_manager =
 		wlr_server_decoration_manager_create(server->wl_display);
 	wlr_server_decoration_manager_set_default_mode(
@@ -172,10 +174,12 @@ void server_fini(struct sway_server *server) {
 
 bool server_start(struct sway_server *server) {
 #if HAVE_XWAYLAND
-	if (config->xwayland) {
-		sway_log(SWAY_DEBUG, "Initializing Xwayland");
+	if (config->xwayland != XWAYLAND_MODE_DISABLED) {
+		sway_log(SWAY_DEBUG, "Initializing Xwayland (lazy=%d)",
+				config->xwayland == XWAYLAND_MODE_LAZY);
 		server->xwayland.wlr_xwayland =
-			wlr_xwayland_create(server->wl_display, server->compositor, true);
+			wlr_xwayland_create(server->wl_display, server->compositor,
+					config->xwayland == XWAYLAND_MODE_LAZY);
 		wl_signal_add(&server->xwayland.wlr_xwayland->events.new_surface,
 			&server->xwayland_surface);
 		server->xwayland_surface.notify = handle_xwayland_surface;
