@@ -6,8 +6,6 @@
 #include <strings.h>
 #include <wayland-server-core.h>
 #include <wlr/types/wlr_output_layout.h>
-#include <wlr/types/wlr_xdg_shell_v6.h>
-#include <wlr/types/wlr_xdg_shell.h>
 #include "cairo.h"
 #include "pango.h"
 #include "sway/config.h"
@@ -187,11 +185,6 @@ static struct sway_container *surface_at_view(struct sway_container *con, double
 				view_sx, view_sy, &_sx, &_sy);
 		break;
 #endif
-	case SWAY_VIEW_XDG_SHELL_V6:
-		_surface = wlr_xdg_surface_v6_surface_at(
-				view->wlr_xdg_surface_v6,
-				view_sx, view_sy, &_sx, &_sy);
-		break;
 	case SWAY_VIEW_XDG_SHELL:
 		_surface = wlr_xdg_surface_surface_at(
 				view->wlr_xdg_surface,
@@ -358,19 +351,6 @@ static bool surface_is_popup(struct wlr_surface *surface) {
 		return false;
 	}
 
-	if (wlr_surface_is_xdg_surface_v6(surface)) {
-		struct wlr_xdg_surface_v6 *xdg_surface_v6 =
-			wlr_xdg_surface_v6_from_wlr_surface(surface);
-		while (xdg_surface_v6 &&
-				xdg_surface_v6->role != WLR_XDG_SURFACE_V6_ROLE_NONE) {
-			if (xdg_surface_v6->role == WLR_XDG_SURFACE_V6_ROLE_POPUP) {
-				return true;
-			}
-			xdg_surface_v6 = xdg_surface_v6->toplevel->parent;
-		}
-		return false;
-	}
-
 	return false;
 }
 
@@ -978,6 +958,11 @@ static void container_fullscreen_workspace(struct sway_container *con) {
 			focus_ws = seat_get_focused_workspace(seat);
 			if (focus_ws == con->workspace) {
 				seat_set_focus_container(seat, con);
+			} else {
+				struct sway_node *focus =
+					seat_get_focus_inactive(seat, &root->node);
+				seat_set_raw_focus(seat, &con->node);
+				seat_set_raw_focus(seat, focus);
 			}
 		}
 	}
@@ -1542,3 +1527,10 @@ void container_raise_floating(struct sway_container *con) {
 bool container_is_scratchpad_hidden(struct sway_container *con) {
 	return con->scratchpad && !con->workspace;
 }
+
+bool container_is_scratchpad_hidden_or_child(struct sway_container *con) {
+	while (con->parent) {
+		con = con->parent;
+	}
+	return con->scratchpad && !con->workspace;
+}
