@@ -60,7 +60,7 @@ static void restore_workspaces(struct sway_output *output) {
 		struct sway_workspace *ws = root->noop_output->workspaces->items[0];
 		workspace_detach(ws);
 		output_add_workspace(output, ws);
-		
+
 		// If the floater was made floating while on the NOOP output, its width
 		// and height will be zero and it should be reinitialized as a floating
 		// container to get the appropriate size and location. Additionally, if
@@ -93,6 +93,7 @@ struct sway_output *output_create(struct wlr_output *wlr_output) {
 	output->wlr_output = wlr_output;
 	wlr_output->data = output;
 	output->detected_subpixel = wlr_output->subpixel;
+	output->scale_filter = SCALE_FILTER_NEAREST;
 
 	wl_signal_init(&output->events.destroy);
 
@@ -104,9 +105,9 @@ struct sway_output *output_create(struct wlr_output *wlr_output) {
 	return output;
 }
 
-void output_enable(struct sway_output *output, struct output_config *oc) {
+bool output_enable(struct sway_output *output, struct output_config *oc) {
 	if (!sway_assert(!output->enabled, "output is already enabled")) {
-		return;
+		return false;
 	}
 	struct wlr_output *wlr_output = output->wlr_output;
 	size_t len = sizeof(output->layers) / sizeof(output->layers[0]);
@@ -117,7 +118,7 @@ void output_enable(struct sway_output *output, struct output_config *oc) {
 	output->enabled = true;
 	if (!apply_output_config(oc, output)) {
 		output->enabled = false;
-		return;
+		return false;
 	}
 
 	output->configured = true;
@@ -155,6 +156,8 @@ void output_enable(struct sway_output *output, struct output_config *oc) {
 
 	arrange_layers(output);
 	arrange_root();
+
+	return true;
 }
 
 static void evacuate_sticky(struct sway_workspace *old_ws,
@@ -240,6 +243,7 @@ void output_destroy(struct sway_output *output) {
 	}
 	list_free(output->workspaces);
 	list_free(output->current.workspaces);
+	wl_event_source_remove(output->repaint_timer);
 	free(output);
 }
 
@@ -267,6 +271,7 @@ void output_disable(struct sway_output *output) {
 
 	output->enabled = false;
 	output->configured = false;
+	output->current_mode = NULL;
 
 	arrange_root();
 }
