@@ -132,8 +132,6 @@ uint32_t view_get_window_type(struct sway_view *view) {
 
 const char *view_get_shell(struct sway_view *view) {
 	switch(view->type) {
-	case SWAY_VIEW_XDG_SHELL_V6:
-		return "xdg_shell_v6";
 	case SWAY_VIEW_XDG_SHELL:
 		return "xdg_shell";
 #if HAVE_XWAYLAND
@@ -217,8 +215,8 @@ void view_autoconfigure(struct sway_view *view) {
 
 	if (!container_is_floating(con) && ws) {
 
-		bool smart = config->hide_edge_borders == E_SMART ||
-			(config->hide_edge_borders == E_SMART_NO_GAPS &&
+		bool smart = config->hide_edge_borders_smart == ESMART_ON ||
+			(config->hide_edge_borders_smart == ESMART_NO_GAPS &&
 			!gaps_to_edge(view));
 		bool hide_smart = smart && view_is_only_visible(view);
 
@@ -451,6 +449,25 @@ void view_execute_criteria(struct sway_view *view) {
 	list_free(criterias);
 }
 
+static void view_populate_pid(struct sway_view *view) {
+	pid_t pid;
+	switch (view->type) {
+#if HAVE_XWAYLAND
+	case SWAY_VIEW_XWAYLAND:;
+		struct wlr_xwayland_surface *surf =
+			wlr_xwayland_surface_from_wlr_surface(view->surface);
+		pid = surf->pid;
+		break;
+#endif
+	case SWAY_VIEW_XDG_SHELL:;
+		struct wl_client *client =
+			wl_resource_get_client(view->surface->resource);
+		wl_client_get_credentials(client, &pid, NULL, NULL);
+		break;
+	}
+	view->pid = pid;
+}
+
 static struct sway_workspace *select_workspace(struct sway_view *view) {
 	struct sway_seat *seat = input_manager_current_seat();
 
@@ -486,28 +503,12 @@ static struct sway_workspace *select_workspace(struct sway_view *view) {
 	}
 	list_free(criterias);
 	if (ws) {
+		root_remove_workspace_pid(view->pid);
 		return ws;
 	}
 
 	// Check if there's a PID mapping
-	pid_t pid;
-#if HAVE_XWAYLAND
-	if (view->type == SWAY_VIEW_XWAYLAND) {
-		struct wlr_xwayland_surface *surf =
-			wlr_xwayland_surface_from_wlr_surface(view->surface);
-		pid = surf->pid;
-	} else {
-		struct wl_client *client =
-			wl_resource_get_client(view->surface->resource);
-		wl_client_get_credentials(client, &pid, NULL, NULL);
-	}
-#else
-	struct wl_client *client =
-		wl_resource_get_client(view->surface->resource);
-	wl_client_get_credentials(client, &pid, NULL, NULL);
-#endif
-	view->pid = pid;
-	ws = root_workspace_for_pid(pid);
+	ws = root_workspace_for_pid(view->pid);
 	if (ws) {
 		return ws;
 	}
@@ -565,6 +566,8 @@ void view_map(struct sway_view *view, struct wlr_surface *wlr_surface,
 		return;
 	}
 	view->surface = wlr_surface;
+	view_populate_pid(view);
+	view->container = container_create(view);
 
 	// If there is a request to be opened fullscreen on a specific output, try
 	// to honor that request. Otherwise, fallback to assigns, pid mappings,
@@ -591,7 +594,6 @@ void view_map(struct sway_view *view, struct wlr_surface *wlr_surface,
 		ws = seat_get_last_known_workspace(seat);
 	}
 
-	view->container = container_create(view);
 	if (target_sibling) {
 		container_add_sibling(target_sibling, view->container, 1);
 	} else if (ws) {
@@ -933,11 +935,6 @@ struct sway_view *view_from_wlr_surface(struct wlr_surface *wlr_surface) {
 			wlr_xdg_surface_from_wlr_surface(wlr_surface);
 		return view_from_wlr_xdg_surface(xdg_surface);
 	}
-	if (wlr_surface_is_xdg_surface_v6(wlr_surface)) {
-		struct wlr_xdg_surface_v6 *xdg_surface_v6 =
-			wlr_xdg_surface_v6_from_wlr_surface(wlr_surface);
-		return view_from_wlr_xdg_surface_v6(xdg_surface_v6);
-	}
 #if HAVE_XWAYLAND
 	if (wlr_surface_is_xwayland_surface(wlr_surface)) {
 		struct wlr_xwayland_surface *xsurface =
